import Graphics.UI.GLUT
import Graphics.Rendering.OpenGL
import Data.IORef


main :: IO ()
main = do
  getArgsAndInitialize
  initialDisplayMode $= [WithDepthBuffer]
  createWindow "The edge of the World"
  windowSize $= Size 800 800
  depthFunc $= Just Less
  angle <- newIORef 0.0				--creating a new mutable variable
  loc <- newIORef (0.0)				--creating a new mutable variable
  displayCallback $= display angle loc		--calling the function that draws the screen
  idleCallback $= Just (idle angle loc)		--calling the function that updates the screen
  mainLoop


display ::IORef GLfloat ->IORef GLfloat -> DisplayCallback
display angle loc  = do
  clear [ColorBuffer,DepthBuffer]
  
  preservingMatrix (do
    translate (Vector3 0 0 (0.9::GLfloat))
    renderPrimitive Polygon screen				--draws the background. A square with different colour vertices
    )
  
  preservingMatrix (do
    translate (Vector3 0.25 (-0.325) (0::GLfloat))
    rotate (185) (Vector3 (-1) (1) ((1)::GLfloat))
    redHolesurf 0.9					--draw the surface of the red hole
    redHole 1						--and now draw the red hole!
    )


  pointSize $= 1
  currentColor $= Color4 1 1 0 1
  render Points (randomPoints 90)  			--draws the stars

  lighting		$= Enabled			--enables haskell's lighting effects. Used to light my planet.
  light (Light 0)	$= Enabled
  position (Light 0)	$= Vertex4 (1) (-1) (0.3) 1	--light from the hole's corner
  ambient (Light 0)	$= Color4 0 0 1 1
  diffuse (Light 0)	$= Color4 0.4 0.4 0.8 1
  attenuation (Light 0)	$=  (0.1, 0.1, (0.1::GLfloat))
  specular (Light 0)	$= Color4 (0.5) 0.5 0.4 1
  
  loadIdentity
  a <- get angle					--my alternating variable. Used to animate the rotation.
  rotate (a) (Vector3 (-0.3) (0.4) ((0)::GLfloat))
  preservingMatrix (do
    translate (Vector3 (-0.3) (0.4) (0::GLfloat))	--rotates the planet around its axis
    currentColor $= Color4 0.2 0.2 1 1
    renderQuadric (QuadricStyle (Just Smooth) GenerateTextureCoordinates Outside FillStyle) (Sphere 0.3 800 800))
  
  lighting  $= Disabled
  light (Light 0) $= Disabled
  
  l <- get loc
  preservingMatrix (do
    pointSize $= 2.5
    currentColor $= Color4 1 1 1 0.1
    loadIdentity
    translate (Vector3 (l) (l) (0.5::GLfloat))
    renderPrimitive Points (fallingstars)  			--draws the falling stars
    )
  
  
  currentColor $= Color4 0.9 0.5 0 0			--my planet's ring. Just a circle...
  translate (Vector3 (-0.3) (0.4) (0::GLfloat))
  render Polygon (circlePoints 0.42 900)
  loadIdentity  
  swapBuffers

--a function that draws the given figure
render:: PrimitiveMode -> [(GLfloat, GLfloat, GLfloat)] -> IO ()
render figure ps = do
  renderPrimitive figure (makePoints ps)

--a function that takes triples and converts them to points on the screen
makePoints :: [(GLfloat, GLfloat, GLfloat)] -> IO ()
makePoints ps = mapM_ (\(x,y,z) -> vertex (Vertex3 (x::GLfloat) y z)) ps

--a function that creates points of a circle's circumference
circlePoints :: GLfloat -> GLfloat -> [(GLfloat,GLfloat,GLfloat)]
circlePoints r n = [(r*sin(i*2*pi / n),r*cos(i*2*pi / n),0) | i <- [0,1..n] , n >= 0]


--the red hole consists of (a lot of) 2D circles, each one in close proximity (in the 3rd dimension) to its next one, and all of them rotated in
-- such an angle that gives the impression (to me at least!) of a 3D whirl. 

redHole:: GLfloat -> IO ()
redHole m = sequence_ [(currentColor $= Color4 (0.0+1.2*n) (0.0) (0.0) 0.7) >> (translate (Vector3 0 0 ((0-(n/8))::GLfloat))) >> (render Polygon (circlePoints (0.7-1.38*n) 1500)) | n <- [0.0,0.01..0.4]]


--the surface of my red hole consists of circles of gradually alternating (red) color

redHolesurf:: GLfloat -> IO ()
redHolesurf m = sequence_ [(currentColor $= Color4 (0.015+n) (0.015) (0.015) 0.1) >> (render Polygon (circlePoints (0.7-1.38*n) 1500)) | n <- [0.0,0.1..m]]


--using trigonometry functions to produce 'random' points ranging from -1 - 1 (so they fit in the screen)!!
randomPoints :: GLfloat -> [(GLfloat,GLfloat,GLfloat)]
randomPoints n = [ (sin (cos (77* x)), -(sin (tan (30*x))), 0) | x <- [1..n] ]


--the positions of my falling stars!
fallingstars :: IO ()
fallingstars = do
  currentColor $= Color4 1 1 0 1
  vertex$Vertex3 (2.2) (2.2) ((0)::GLfloat)
  vertex$Vertex3 (0.3) (1.2) (0::GLfloat)
  vertex$Vertex3 (-0.3) (1.3) (0::GLfloat)
  vertex$Vertex3 (1.2) (0.3) (0::GLfloat)
  vertex$Vertex3 (-0.5) (1.8) (0::GLfloat)
  vertex$Vertex3 (0) (1) (0::GLfloat)

  
--my screen's corners and their colors
screen :: IO ()
screen = do
  currentColor $= Color4 0.1 0 0.1 1
  vertex$Vertex3 (1) (1) (0::GLfloat)
  
  currentColor $= Color4 0.3 0.2 0.6 1
  vertex$Vertex3 (1) (-1) (0::GLfloat)
  
  currentColor $= Color4 0.1 0.1 0 1
  vertex$Vertex3 (-1) (-1) (0::GLfloat)
  
  currentColor $= Color4 0.2 0.1 0.1 1
  vertex$Vertex3 (-1) (1) (0::GLfloat)

  
--The functions that updates the rotation angle
idle :: IORef GLfloat ->IORef GLfloat -> IdleCallback
idle angle loc  = do
  angle $~! (+ 3)	--the angle mutable state variable gets increased by 3 each time
  l <- get loc
  if (l <= -3.5 ) then loc $= 0.0 		--if new star position gets too far under the screen, reset it
		else loc $= (l-0.035)
  postRedisplay Nothing